"""Classes relating to planestress geometry."""

from __future__ import annotations

import warnings
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any

import numpy as np
from shapely import GeometryCollection, LineString, MultiPolygon, Polygon, affinity
from triangle import triangulate

import planestress.pre.boundary_condition as bc
from planestress.post.plotting import plotting_context
from planestress.pre.material import DEFAULT_MATERIAL, Material
from planestress.pre.mesh import Mesh


if TYPE_CHECKING:
    import matplotlib.axes


class Geometry:
    """Class describing a geometric region."""

    def __init__(
        self,
        polygons: Polygon | MultiPolygon,
        materials: Material | list[Material] = DEFAULT_MATERIAL,
        tol: int = 12,
    ) -> None:
        """Inits the Geometry class.

        Note ensure length of materials = number of polygons
        """
        # convert polygon to multipolygon
        if isinstance(polygons, Polygon):
            polygons = MultiPolygon(polygons=[polygons])

        # convert material to list of materials
        if isinstance(materials, Material):
            materials = [materials] * len(polygons.geoms)

        # check materials length
        if len(polygons.geoms) != len(materials):
            raise ValueError(
                f"Length of materials: {len(materials)}, must equal number of polygons:"
                f"{len(polygons.geoms)}."
            )

        # save input data
        self.polygons = polygons
        self.materials = materials
        self.tol = tol

        # allocate points, facets, holes, control_points
        self.points: list[Point] = []
        self.facets: list[Facet] = []
        self.holes: list[Point] = []
        self.control_points: list[Point] = []

        # compile the geometry into points, facets, holes and control_points
        self.compile_geometry()

        # allocate mesh variables
        self.mesh: Mesh | None = None
        self.point_markers = [0] * len(self.points)
        self.facet_markers = [0] * len(self.facets)

    def compile_geometry(self) -> None:
        """Creates points, facets, holes and control_points from shapely geometry."""
        # loop through each polygon
        for poly in self.polygons.geoms:
            # first create points,facets, holes and control_points for each polygon
            poly_points, poly_facets, poly_holes, poly_cp = self.compile_polygon(
                polygon=poly
            )

            # add points to the global list, skipping duplicate points
            for pt in poly_points:
                if pt not in self.points:
                    self.points.append(pt)
                # we have to fix facets that reference the skipped point!
                else:
                    # get the point that we are keeping in the list
                    kept_pt_idx = self.points.index(pt)
                    kept_pt = self.points[kept_pt_idx]

                    # loop through all facets and update removed point reference
                    for fct in poly_facets:
                        fct.update_point(old=pt, new=kept_pt)

            # add facets to the global list
            self.facets.extend(poly_facets)

            # add holes to list of multipolygon holes
            self.holes.extend(poly_holes)

            # add control points to the global list
            self.control_points.append(poly_cp)

        # assign point indices
        for idx, pt in enumerate(self.points):
            pt.idx = idx

    def compile_polygon(
        self,
        polygon: Polygon,
    ) -> tuple[list[Point], list[Facet], list[Point], Point]:
        """Create a list of points, facets and holes + control point given a Polygon."""
        pt_list: list[Point] = []
        fct_list: list[Facet] = []
        hl_list: list[Point] = []

        # construct perimeter points (note shapely doubles up first & last point)
        for coords in list(polygon.exterior.coords[:-1]):
            new_pt = Point(x=coords[0], y=coords[1], tol=self.tol)
            pt_list.append(new_pt)

        # create perimeter facets
        fct_list.extend(self.create_facet_list(pt_list=pt_list))

        # construct holes, for each interior (hole) region
        for hl in polygon.interiors:
            int_pt_list: list[Point] = []

            # create hole (note shapely doubles up first & last point)
            for coords in hl.coords[:-1]:
                new_pt = Point(x=coords[0], y=coords[1], tol=self.tol)
                int_pt_list.append(new_pt)

            # add interior points to poly list
            pt_list.extend(int_pt_list)

            # create hole facets
            fct_list.extend(self.create_facet_list(pt_list=int_pt_list))

            # create hole point
            # first convert the list of interior points to a list of tuples
            int_pt_list_tup = [hl_pt.to_tuple() for hl_pt in int_pt_list]

            # create a polygon of the hole region
            hl_poly = Polygon(int_pt_list_tup)

            # add hole point to the list of hole points
            hl_pt_coords = hl_poly.representative_point().coords
            hl_list.append(
                Point(x=hl_pt_coords[0][0], y=hl_pt_coords[0][1], tol=self.tol)
            )

        # construct control point
        cp_pt_coords = polygon.representative_point().coords
        cp_pt = Point(x=cp_pt_coords[0][0], y=cp_pt_coords[0][1], tol=self.tol)

        return pt_list, fct_list, hl_list, cp_pt

    @staticmethod
    def create_facet_list(pt_list: list[Point]) -> list[Facet]:
        """Creates a closed list of facets from a list of points."""
        fct_list: list[Facet] = []

        # create facets
        for idx, pt in enumerate(pt_list):
            pt1 = pt
            # if we are not at the end of the list
            if idx + 1 != len(pt_list):
                pt2 = pt_list[idx + 1]
            # otherwise loop back to starting point
            else:
                pt2 = pt_list[0]

            fct_list.append(Facet(pt1=pt1, pt2=pt2))

        return fct_list

    def align_to(
        self,
        other: Geometry | tuple[float, float],
        on: str,
        inner: bool = False,
    ) -> Geometry:
        """Aligns the geometry to another ``Geometry`` or point."""
        # setup mappings for transformations
        align_self_map = {
            "left": 1,
            "right": 0,
            "bottom": 3,
            "top": 2,
        }
        other_as_geom_map = {
            "left": 0,
            "right": 1,
            "bottom": 2,
            "top": 3,
        }
        other_as_point_map = {
            "left": 0,
            "right": 0,
            "bottom": 1,
            "top": 1,
        }

        # get the coordinate to align from
        self_extents = self.calculate_extents()
        self_align_idx = align_self_map[on]
        self_align_coord = self_extents[self_align_idx]

        # get the coordinate to align to
        if isinstance(other, Geometry):
            align_to_idx = other_as_geom_map[on]
            align_to_coord = other.calculate_extents()[align_to_idx]
        else:
            align_to_idx = other_as_point_map[on]
            align_to_coord = other[align_to_idx]

        if inner:
            self_align_coord = self_extents[align_to_idx]

        # calculate offset
        offset = align_to_coord - self_align_coord

        # shift section
        if on in ["top", "bottom"]:
            arg = "y"
        else:
            arg = "x"

        kwargs = {arg: offset}

        return self.shift_section(**kwargs)

    def align_centre(
        self,
        align_to: Geometry | tuple[float, float] | None = None,
    ) -> Geometry:
        """Aligns the geometry to a centre point."""
        cx, cy = self.calculate_centroid()

        # align to geometry centroid
        if align_to is None:
            shift_x, shift_y = round(-cx, self.tol), round(-cy, self.tol)
        # align to centroid of another geometry
        elif isinstance(align_to, Geometry):
            other_cx, other_cy = align_to.calculate_centroid()
            shift_x = round(other_cx - cx, self.tol)
            shift_y = round(other_cy - cy, self.tol)
        # align to a point
        else:
            try:
                point_x, point_y = align_to
                shift_x = round(point_x - cx, self.tol)
                shift_y = round(point_y - cy, self.tol)
            except (TypeError, ValueError) as exc:
                raise ValueError(
                    f"align_to must be either a Geometry object or an (x, y) "
                    f"coordinate, not {align_to}."
                ) from exc

        return self.shift_section(x=shift_x, y=shift_y)

    def shift_section(
        self,
        x: float = 0.0,
        y: float = 0.0,
    ) -> Geometry:
        """Shifts the geometry by (``x``, ``y``)."""
        return Geometry(
            polygons=affinity.translate(geom=self.polygons, xoff=x, yoff=y),
            materials=self.materials,
            tol=self.tol,
        )

    def rotate_section(
        self,
        angle: float,
        rot_point: tuple[float, float] | str = "center",
        use_radians: bool = False,
    ) -> Geometry:
        """Rotates the geometry."""
        return Geometry(
            polygons=affinity.rotate(
                geom=self.polygons,
                angle=angle,
                origin=rot_point,
                use_radians=use_radians,
            ),
            materials=self.materials,
            tol=self.tol,
        )

    def mirror_section(
        self,
        axis: str = "x",
        mirror_point: tuple[float, float] | str = "center",
    ) -> Geometry:
        """Compound & geom."""
        if axis == "x":
            xfact = 1.0
            yfact = -1.0
        elif axis == "y":
            xfact = -1.0
            yfact = 1.0
        else:
            raise ValueError(f"axis must be 'x' or 'y', not {axis}.")

        return Geometry(
            polygons=affinity.scale(
                geom=self.polygons, xfact=xfact, yfact=yfact, origin=mirror_point
            ),
            materials=self.materials,
            tol=self.tol,
        )

    def calculate_area(self) -> float:
        """Calculates the area of the geometry.

        Returns:
            Geometry area
        """
        return float(self.polygons.area)

    def calculate_centroid(self) -> tuple[float, float]:
        """Calculates the centroid of the geometry.

        Returns:
            Geometry centroid
        """
        x, y = self.polygons.centroid.coords[0]

        return float(x), float(y)

    def calculate_extents(self) -> tuple[float, float, float, float]:
        """Calculate geometry extents.

        Calculates the minimum and maximum ``x`` and ``y`` values amongst the list of
        points, i.e. the points that describe the bounding box of the ``Geometry``
        instance.

        Returns:
            Minimum and maximum ``x`` and ``y`` values (``x_min``, ``x_max``, ``y_min``,
            ``y_max``)
        """
        min_x, min_y, max_x, max_y = self.polygons.bounds
        return min_x, max_x, min_y, max_y

    def __or__(
        self,
        other: Geometry,
    ) -> Geometry:
        """Performs a difference operation using the ``|`` operator.

        Note - applies material of first geometry to entire new geom.
        """
        try:
            new_polygon = self.filter_non_polygons(
                input_geom=self.polygons | other.polygons
            )

            return Geometry(
                polygons=new_polygon, materials=self.materials[0], tol=self.tol
            )
        except Exception as exc:
            raise ValueError(
                f"Cannot perform union on these two objects: {self} | {other}"
            ) from exc

    def __sub__(
        self,
        other: Geometry,
    ) -> Geometry:
        """Performs a difference operation using the ``-`` operator.

        Note about combining resulting geometry into one section! Check in materials!
        """
        try:
            new_polygon = self.filter_non_polygons(
                input_geom=self.polygons - other.polygons
            )

            # non-polygon results
            if isinstance(new_polygon, GeometryCollection):
                raise ValueError(
                    f"Cannot perform difference on these two objects: {self} - {other}"
                )
            # polygon or multipolygon object
            elif isinstance(new_polygon, (Polygon, MultiPolygon)):
                return Geometry(
                    polygons=new_polygon, materials=self.materials, tol=self.tol
                )
            else:
                raise ValueError(
                    f"Cannot perform difference on these two objects: {self} - {other}"
                )
        except Exception as exc:
            raise ValueError(
                f"Cannot perform difference on these two objects: {self} - {other}"
            ) from exc

    def __add__(
        self,
        other: Geometry,
    ) -> Geometry:
        """Performs an addition operation using the ``+`` operator.

        Keeps the largest tol.
        """
        poly_list: list[Polygon] = []
        mat_list: list[Material] = []

        # loop through each list of polygons and combine
        for poly, mat in zip(self.polygons.geoms, self.materials):
            poly_list.append(poly)
            mat_list.append(mat)

        for poly, mat in zip(other.polygons.geoms, other.materials):
            poly_list.append(poly)
            mat_list.append(mat)

        tol = max(self.tol, other.tol)

        return Geometry(
            polygons=MultiPolygon(polygons=poly_list), materials=mat_list, tol=tol
        )

    def __and__(
        self,
        other: Geometry,
    ) -> Geometry:
        """Performs an intersection operation using the ``&`` operator.

        Note - applies material of first geometry to entire new geom.
        """
        try:
            new_polygon = self.filter_non_polygons(
                input_geom=self.polygons - other.polygons
            )

            return Geometry(
                polygons=new_polygon, materials=self.materials[0], tol=self.tol
            )
        except Exception as exc:
            raise ValueError(
                f"Cannot perform intersection on these two Geometry instances: "
                f"{self} & {other}"
            ) from exc

    @staticmethod
    def filter_non_polygons(
        input_geom: GeometryCollection | LineString | Point | Polygon | MultiPolygon,
    ) -> Polygon | MultiPolygon:
        """Filters shapely geometries to return only polygons.

        Returns a ``Polygon`` or a ``MultiPolygon`` representing any such ``Polygon`` or
        ``MultiPolygon`` that may exist in the ``input_geom``. If ``input_geom`` is a
        ``LineString`` or ``Point``, an empty ``Polygon`` is returned.

        Args:
            input_geom: Shapely geometry to filter

        Returns:
            Filtered polygon
        """
        if isinstance(input_geom, (Polygon, MultiPolygon)):
            return input_geom
        elif isinstance(input_geom, GeometryCollection):
            acc = []

            for item in input_geom.geoms:
                if isinstance(item, (Polygon, MultiPolygon)):
                    acc.append(item)

            if len(acc) == 0:
                return Polygon()
            elif len(acc) == 1:
                return acc[0]
            else:
                return MultiPolygon(polygons=acc)
        elif isinstance(input_geom, (Point, LineString)):
            return Polygon()
        else:
            return Polygon()

    def add_node_support(
        self,
        point: tuple[float, float],
        direction: str,
        value: float = 0.0,
    ) -> bc.NodeSupport:
        """Adds a node support to the geometry."""
        # check to see if a mesh has already been generated
        self.check_boundary_condition_mesh()

        # create point object
        pt = Point(point[0], point[1], tol=self.tol)

        # find point in list of points
        try:
            idx = self.points.index(pt)
        except ValueError as exc:
            raise ValueError("Cannot find the point: {pt} in the geometry.") from exc

        # check to see if point already has custom point marker
        if self.point_markers[idx] == 0:
            # add point marker to point, note custom point marker ids start at 2
            marker_id = max(1, max(self.point_markers)) + 1
            self.point_markers[idx] = marker_id
        else:
            # get marker id
            marker_id = self.point_markers[idx]

        # create node support boundary condition
        ns_bc = bc.NodeSupport(marker_id=marker_id, direction=direction, value=value)

        return ns_bc

    def check_boundary_condition_mesh(self) -> None:
        """Checks to see if a mesh exist before creating a new boundary condition.

        If a mesh exists, warns the user to regenerate the mesh prior to conducting an
        analysis.
        """
        if self.mesh:
            warnings.warn(
                "Please regenerate the mesh prior to creating a PlaneStress object."
            )

    def create_mesh(
        self,
        mesh_sizes: float | list[float] = 0.0,
        linear: bool = True,
        min_angle: float = 30.0,
        coarse: bool = False,
    ) -> None:
        """Creates a triangular mesh of the geometry.

        Args:
            mesh_sizes: A float describing the maximum mesh element area to be used in
                the finite-element mesh for each polygon the ``Geometry`` object. If a
                list of length 1 or a ``float`` is passed, then the one size will be
                applied to all polygons. A value of ``0`` removes the area constraint.
            linear: Order of triangular mesh, if ``True`` gives linear elements and if
                ``False`` gives quadratic elements
            min_angle: The meshing algorithm adds vertices to the mesh to ensure that no
                angle smaller than the minimum angle (in degrees, rounded to 1 decimal
                place). Note that small angles between input segments cannot be
                eliminated. If the minimum angle is 20.7 deg or smaller, the
                triangulation algorithm is theoretically guaranteed to terminate (given
                sufficient precision). The algorithm often doesn't terminate for angles
                greater than 33 deg. Some meshes may require angles well below 20 deg to
                avoid problems associated with insufficient floating-point precision.
            coarse: If set to True, will create a coarse mesh (no area or quality
                constraints)
        """
        if isinstance(mesh_sizes, (float, int)):
            mesh_sizes = [mesh_sizes]

        if len(mesh_sizes) == 1:
            mesh_sizes = mesh_sizes * len(self.control_points)

        tri: dict[str, Any] = {}  # create tri dictionary
        tri["vertices"] = [pt.to_tuple() for pt in self.points]  # set points
        tri["vertex_markers"] = self.point_markers  # set point markers
        tri["segments"] = [fct.to_tuple() for fct in self.facets]  # set facets

        if self.holes:
            tri["holes"] = [pt.to_tuple() for pt in self.holes]  # set holes

        # prepare regions
        regions: list[list[float | int]] = []

        for idx, cp in enumerate(self.control_points):
            rg = [cp.x, cp.y, idx, mesh_sizes[idx]]
            regions.append(rg)

        tri["regions"] = regions  # set regions

        # set mesh order
        m_order = "" if linear else "o2"

        # generate mesh
        if coarse:
            mesh = triangulate(tri, "pA{m_order}")
        else:
            mesh = triangulate(tri, f"pq{min_angle:.1f}Aa{m_order}")

        self.mesh = Mesh(
            nodes=np.array(mesh["vertices"], dtype=np.float64),
            elements=np.array(mesh["triangles"], dtype=np.int32),
            attributes=np.array(
                mesh["triangle_attributes"].T[0], dtype=np.int32
            ).tolist(),
            node_markers=np.array(mesh["vertex_markers"].T[0], dtype=np.int32).tolist(),
            linear=linear,
        )

        # re-order mid-nodes if quadratic
        if not linear:
            self.mesh.elements[:, [3, 4, 5]] = self.mesh.elements[:, [5, 3, 4]]

    def plot_geometry(
        self,
        labels: tuple[str] = ("control_points",),
        title: str = "Geometry",
        cp: bool = True,
        legend: bool = True,
        **kwargs: Any,
    ) -> matplotlib.axes.Axes:
        """Plots the geometry."""
        # create plot and setup the plot
        with plotting_context(title=title, **kwargs) as (_, ax):
            assert ax
            label: str | None

            # plot the points and facets
            label = "Points & Facets"
            for fct in self.facets:
                ax.plot(
                    (fct.pt1.x, fct.pt2.x),
                    (fct.pt1.y, fct.pt2.y),
                    "ko-",
                    markersize=2,
                    linewidth=1.5,
                    label=label,
                )
                label = None

            # plot the holes
            label = "Holes"
            for hl in self.holes:
                ax.plot(hl.x, hl.y, "rx", markersize=5, markeredgewidth=1, label=label)
                label = None

            # plot the control points
            if cp:
                label = "Control Points"
                for cpts in self.control_points:
                    ax.plot(cpts.x, cpts.y, "bo", markersize=5, label=label)
                    label = None

            # display the labels
            for label in labels:
                # plot control_point labels
                if label == "control_points":
                    for idx, pt in enumerate(self.control_points):
                        ax.annotate(str(idx), xy=(pt.x, pt.y), color="b")

                # plot point labels
                if label == "points":
                    for idx, pt in enumerate(self.points):
                        ax.annotate(str(idx), xy=(pt.x, pt.y), color="r")

                # plot facet labels
                if label == "facets":
                    for idx, fct in enumerate(self.facets):
                        xy = (fct.pt1.x + fct.pt2.x) / 2, (fct.pt1.y + fct.pt2.y) / 2

                        ax.annotate(str(idx), xy=xy, color="b")

                # plot hole labels
                if label == "holes":
                    for idx, pt in enumerate(self.holes):
                        ax.annotate(str(idx), xy=(pt.x, pt.y), color="r")

            # display the legend
            if legend:
                ax.legend(loc="center left", bbox_to_anchor=(1, 0.5))

        return ax

    def plot_mesh(
        self,
        nodes: bool = False,
        nd_num: bool = False,
        el_num: bool = False,
        materials: bool = False,
        alpha: float = 0.5,
        mask: list[bool] | None = None,
        title: str = "Finite Element Mesh",
        **kwargs: Any,
    ) -> matplotlib.axes.Axes:
        """Plots the finite element mesh."""
        if self.mesh is not None:
            return self.mesh.plot_mesh(
                material_list=self.materials,
                nodes=nodes,
                nd_num=nd_num,
                el_num=el_num,
                materials=materials,
                alpha=alpha,
                mask=mask,
                title=title,
                **kwargs,
            )
        else:
            raise RuntimeError("Generate a mesh with create_mesh() prior to plotting.")


@dataclass(eq=True)
class Point:
    """Class describing a point in 2D space."""

    x: float
    y: float
    tol: int
    idx: int | None = None

    def __post_init__(self) -> None:
        """Point object post init method."""
        self.round()

    def __eq__(
        self,
        other: Point,
    ) -> bool:
        """Override __eq__ method to neglect index."""
        return self.x == other.x and self.y == other.y

    def round(self) -> None:
        """Rounds the point to ``tol`` digits."""
        self.x = round(self.x, self.tol)
        self.y = round(self.y, self.tol)

    def to_tuple(self) -> tuple[float, float]:
        """Converts the point to a tuple."""
        return self.x, self.y


@dataclass(eq=True)
class Facet:
    """Class describing a facet of a 2D geometry, i.e. an edge."""

    pt1: Point
    pt2: Point

    def to_tuple(self) -> tuple[float, float]:
        """Converts the facet to a tuple."""
        idx_1 = self.pt1.idx
        idx_2 = self.pt2.idx

        if idx_1 is None:
            raise RuntimeError(f"Point 1: {self.pt1} has not been assigned an index.")

        if idx_2 is None:
            raise RuntimeError(f"Point 2: {self.pt2} has not been assigned an index.")

        return idx_1, idx_2

    def update_point(
        self,
        old: Point,
        new: Point,
    ) -> None:
        """If the facet contains the point ``old``, replace with ``new``."""
        if self.pt1 == old:
            self.pt1 = new

        if self.pt2 == old:
            self.pt2 = new



"""Finite element classes for a plane-stress analysis."""

from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np
import numpy.typing as npt


if TYPE_CHECKING:
    from planestress.pre.material import Material

TRI_ARRAY = np.array([[0.0, 1.0, 0.0], [0.0, 0.0, 1.0]])


class FiniteElement:
    """Abstract base class for a triangular plane-stress finite element."""

    def __init__(
        self,
        el_idx: int,
        coords: npt.NDArray[np.float64],
        node_idxs: list[int],
        material: Material,
    ) -> None:
        """Inits the FiniteElement class."""
        self.el_idx = el_idx
        self.coords = coords
        self.node_idxs = node_idxs
        self.material = material
        self.num_nodes: int = 0

    def __str__(self) -> str:
        """Override string method."""
        return (
            f"{self.__class__.__name__} - id: {self.el_idx}, material: "
            f"{self.material.name}"
        )

    def shape_functions(
        self,
        gauss_point: tuple[float, float, float],
    ) -> npt.NDArray[np.float64]:
        """Returns the shape functions for the finite element."""
        raise NotImplementedError

    def shape_functions_derivatives(
        self,
        gauss_point: tuple[float, float, float],
    ) -> npt.NDArray[np.float64]:
        """Returns the derivatives of the shape functions for the finite element."""
        raise NotImplementedError

    def global_gauss_point(
        self,
        gauss_point: tuple[float, float, float],
    ) -> tuple[float, float]:
        """Returns the global coordinates of the gauss point."""
        nx, ny = self.coords @ self.shape_functions(gauss_point=gauss_point)

        return nx, ny

    def b_matrix_jacobian(
        self,
        gauss_point: tuple[float, float, float],
    ) -> tuple[npt.NDArray[np.float64], float]:
        """Calculates properties related to the finite element.

        See Felippa Ch. 24
        """
        b_iso = self.shape_functions_derivatives(gauss_point=gauss_point)

        # form Jacobian matrix
        j = np.ones((3, 3))
        j[:, 1:] = b_iso @ self.coords.transpose()

        # calculate the jacobian
        jacobian = 0.5 * np.linalg.det(j)

        # if the area of the element is not zero
        if jacobian != 0:
            b_mat = TRI_ARRAY @ np.linalg.solve(j, b_iso)
        else:
            b_mat = np.zeros((2, self.num_nodes))  # empty b matrix

        # form plane stress b matrix
        b_mat_ps = np.zeros((3, 2 * self.num_nodes))

        # fill first two rows with first two rows of b matrix
        # first row - every second entry starting with first
        # second row - every second entry starting with second
        for i in range(2):
            b_mat_ps[i, i::2] = b_mat[i, :]

        # last row:
        # fill every second entry (starting with the first) from second row of b matrix
        b_mat_ps[2, ::2] = b_mat[1, :]
        # fill every second entry (starting with the second) from first row of b matrix
        b_mat_ps[2, 1::2] = b_mat[0, :]

        return b_mat_ps, jacobian

    def element_stiffness_matrix(
        self,
        n_points: int,
    ) -> npt.NDArray[np.float64]:
        """Assembles the stiffness matrix for the element."""
        # allocate element stiffness matrix
        k_el = np.zeros((2 * self.num_nodes, 2 * self.num_nodes))

        # get d_matrix
        d_mat = self.material.get_d_matrix()

        # get Gauss points
        gauss_points = self.gauss_points(n_points=n_points)

        # loop through each gauss point
        for gauss_point in gauss_points:
            b_mat, j = self.b_matrix_jacobian(gauss_point=gauss_point[1:])

            # calculate stiffness matrix for current integration point
            k_el += (
                b_mat.transpose()
                @ d_mat
                @ b_mat
                * gauss_point[0]
                * j
                * self.material.thickness
            )

        return k_el

    def element_load_vector(
        self,
        n_points: int,
    ) -> npt.NDArray[np.float64]:
        """Assembles the load matrix for the element."""
        # allocate element load vector
        k_el = np.zeros(2 * self.num_nodes)

        # TODO - implement!

        return k_el

    @staticmethod
    def gauss_points(n_points: int) -> npt.NDArray[np.float64]:
        """Gaussian weights and locations for ``n_point`` Gaussian integration."""
        # one point gaussian integration
        if n_points == 1:
            return np.array([[1.0, 1.0 / 3, 1.0 / 3, 1.0 / 3]])

        # three point gaussian integration
        if n_points == 3:
            return np.array(
                [
                    [1.0 / 3, 2.0 / 3, 1.0 / 6, 1.0 / 6],
                    [1.0 / 3, 1.0 / 6, 2.0 / 3, 1.0 / 6],
                    [1.0 / 3, 1.0 / 6, 1.0 / 6, 2.0 / 3],
                ]
            )

        # four-point integration
        if n_points == 4:
            return np.array(
                [
                    [-27.0 / 48.0, 1.0 / 3.0, 1.0 / 3.0, 1.0 / 3.0],
                    [25.0 / 48.0, 0.6, 0.2, 0.2],
                    [25.0 / 48.0, 0.2, 0.6, 0.2],
                    [25.0 / 48.0, 0.2, 0.2, 0.6],
                ]
            )

        # six point gaussian integration
        if n_points == 6:
            g1 = 1.0 / 18 * (8 - np.sqrt(10) + np.sqrt(38 - 44 * np.sqrt(2.0 / 5)))
            g2 = 1.0 / 18 * (8 - np.sqrt(10) - np.sqrt(38 - 44 * np.sqrt(2.0 / 5)))
            w1 = (620 + np.sqrt(213125 - 53320 * np.sqrt(10))) / 3720
            w2 = (620 - np.sqrt(213125 - 53320 * np.sqrt(10))) / 3720

            return np.array(
                [
                    [w2, 1 - 2 * g2, g2, g2],
                    [w2, g2, 1 - 2 * g2, g2],
                    [w2, g2, g2, 1 - 2 * g2],
                    [w1, g1, g1, 1 - 2 * g1],
                    [w1, 1 - 2 * g1, g1, g1],
                    [w1, g1, 1 - 2 * g1, g1],
                ]
            )

        raise ValueError("n must be 1, 3, 4 or 6.")


class Tri3(FiniteElement):
    """Class for a three-noded linear triangular element.

    Coords: 2 x 3
    """

    def __init__(
        self,
        el_idx: int,
        coords: npt.NDArray[np.float64],
        node_idxs: list[int],
        material: Material,
    ) -> None:
        """Inits the Tri3 class."""
        super().__init__(
            el_idx=el_idx, coords=coords, node_idxs=node_idxs, material=material
        )
        self.num_nodes: int = 3

    def shape_functions(
        self,
        gauss_point: tuple[float, float, float],
    ) -> npt.NDArray[np.float64]:
        """Returns the shape functions for a Tri3 element."""
        # location of isoparametric co-ordinates for each Gauss point
        eta, xi, zeta = gauss_point

        return np.array([eta, xi, zeta])

    def shape_functions_derivatives(
        self,
        gauss_point: tuple[float, float, float],
    ) -> npt.NDArray[np.float64]:
        """Returns the derivatives of the shape functions for a Tri3 element."""
        # derivatives of the shape functions wrt the isoparametric co-ordinates
        return np.array(
            [
                [1, 0, 0],  # d/d(eta)
                [0, 1, 0],  # d/d(xi)
                [0, 0, 1],  # d/d(zeta)
            ]
        )


class Tri6(FiniteElement):
    """Class for a six-noded quadratic triangular element.

    Coords: 2 x 6
    """

    def __init__(
        self,
        el_idx: int,
        coords: npt.NDArray[np.float64],
        node_idxs: list[int],
        material: Material,
    ) -> None:
        """Inits the Tri6 class."""
        super().__init__(
            el_idx=el_idx, coords=coords, node_idxs=node_idxs, material=material
        )
        self.num_nodes: int = 6

    def shape_functions(
        self,
        gauss_point: tuple[float, float, float],
    ) -> npt.NDArray[np.float64]:
        """Returns the shape functions for a Tri3 element."""
        # location of isoparametric co-ordinates for each Gauss point
        eta, xi, zeta = gauss_point

        return np.array(
            [
                eta * (2 * eta - 1),
                xi * (2 * xi - 1),
                zeta * (2 * zeta - 1),
                4 * eta * xi,
                4 * xi * zeta,
                4 * eta * zeta,
            ],
        )

    def shape_functions_derivatives(
        self,
        gauss_point: tuple[float, float, float],
    ) -> npt.NDArray[np.float64]:
        """Returns the derivatives of the shape functions for a Tri3 element."""
        # location of isoparametric co-ordinates for each Gauss point
        eta, xi, zeta = gauss_point

        # derivatives of the shape functions wrt the isoparametric co-ordinates
        return np.array(
            [
                [4 * eta - 1, 0, 0, 4 * xi, 0, 4 * zeta],  # d/d(eta)
                [0, 4 * xi - 1, 0, 4 * eta, 4 * zeta, 0],  # d/d(xi)
                [0, 0, 4 * zeta - 1, 0, 4 * xi, 4 * eta],  # d/d(zeta)
            ],
        )



"""Class for storing ``planestress`` results."""

from __future__ import annotations

from dataclasses import dataclass

import numpy as np
import numpy.typing as npt


@dataclass
class Results:
    """Class for plane-stress results."""

    u: npt.NDArray[np.float64]



"""planestress post-processor plotting functions."""

from __future__ import annotations

import contextlib
from collections.abc import Generator
from typing import TYPE_CHECKING, Any

import matplotlib.pyplot as plt


if TYPE_CHECKING:
    import matplotlib.axes
    import matplotlib.figure


@contextlib.contextmanager
def plotting_context(
    ax: matplotlib.axes.Axes | None = None,
    pause: bool = True,
    title: str = "",
    filename: str = "",
    render: bool = True,
    axis_index: int | tuple[int, int] | None = None,
    **kwargs: Any,
) -> Generator[
    tuple[matplotlib.figure.Figure, matplotlib.axes.Axes | Any | None], None, None
]:
    """Executes code required to set up a matplotlib figure.

    Args:
        ax: Axes object on which to plot
        pause: If set to True, the figure pauses the script until the window is closed.
            If set to False, the script continues immediately after the window is
            rendered.
        title: Plot title
        filename: Pass a non-empty string or path to save the image as. If this option
            is used, the figure is closed after the file is saved.
        render: If set to False, the image is not displayed. This may be useful if the
            figure or axes will be embedded or further edited before being displayed.
        axis_index: If more than 1 axis is created by subplot, then this is the axis to
            plot on. This may be a tuple if a 2D array of plots is returned.  The
            default value of None will select the top left plot.
        kwargs: Passed to :func:`matplotlib.pyplot.subplots`

    Raises:
        ValueError: ``axis_index`` is invalid

    Yields:
        Matplotlib figure and axes
    """
    if filename:
        render = False

    if ax is None:
        if not render or pause:
            plt.ioff()
        else:
            plt.ion()

        ax_supplied = False
        fig, ax = plt.subplots(**kwargs)

        try:
            if axis_index is None:
                axis_index = (0,) * ax.ndim  # type: ignore

            ax = ax[axis_index]  # type: ignore
        except (AttributeError, TypeError):
            pass  # only 1 axis, not an array
        except IndexError as exc:
            raise ValueError(
                f"axis_index={axis_index} is not compatible with arguments to "
                f"subplots: {kwargs}."
            ) from exc
    else:
        fig = ax.get_figure()  # type: ignore
        ax_supplied = True

        if not render:
            plt.ioff()

    yield fig, ax

    if ax is not None:
        ax.set_title(title)
        plt.tight_layout()
        ax.set_aspect("equal", anchor="C")

    # if no axes was supplied, finish the plot and return the figure and axes
    if ax_supplied:
        # if an axis was supplied, don't continue displaying or configuring the plot
        return

    if filename:
        fig.savefig(filename, dpi=fig.dpi)
        plt.close(fig)  # close the figure to free the memory
        return  # if the figure was to be saved, then don't show it also

    if render:
        if pause:
            plt.show()  # type: ignore
        else:
            plt.draw()
            plt.pause(0.001)



"""Class describing a planestress load case."""

from __future__ import annotations

from dataclasses import dataclass

from planestress.pre.boundary_condition import BoundaryCondition


@dataclass
class LoadCase:
    """Class for a load case."""

    boundary_conditions: list[BoundaryCondition]
    global_accelerations: list[float] | None = None  # TODO - verify list lengths



"""Classes describing a planestress boundary conditions."""

from __future__ import annotations

import numpy as np
import numpy.typing as npt


class BoundaryCondition:
    """Abstract base class for a boundary condition."""

    def __init__(
        self,
        marker_id: int,
    ) -> None:
        """Inits the BoundaryCondition class."""
        self.marker_id = marker_id

    def apply_bc(
        self,
        k: npt.NDArray[np.float64],
        f: npt.NDArray[np.float64],
        dofs: list[int],
    ) -> tuple[npt.NDArray[np.float64], npt.NDArray[np.float64]]:
        """Applies the boundary condition."""
        raise NotImplementedError


class NodeBoundaryCondition(BoundaryCondition):
    """Abstract base class for a boundary condition at a node."""

    def __init__(
        self,
        marker_id: int,
        direction: str,
        value: float,
    ) -> None:
        """Inits the NodeBoundaryCondition class."""
        super().__init__(marker_id=marker_id)
        self.direction = direction  # TODO - verify
        self.value = value


class NodeSupport(NodeBoundaryCondition):
    """Class for adding a support to a node."""

    def __init__(
        self,
        marker_id: int,
        direction: str,
        value: float,
    ) -> None:
        """Inits the NodeSupport class."""
        super().__init__(marker_id=marker_id, direction=direction, value=value)

    def apply_bc(
        self,
        k: npt.NDArray[np.float64],
        f: npt.NDArray[np.float64],
        dofs: list[int],
    ) -> tuple[npt.NDArray[np.float64], npt.NDArray[np.float64]]:
        """Applies the boundary condition."""
        # get relevant dof
        dof = dofs[0] if self.direction == "x" else dofs[1]

        # apply bc
        k[dof, :] = 0
        k[dof, dof] = 1
        f[dof] = self.value

        return k, f


class NodeSpring(NodeBoundaryCondition):
    """Class for adding a spring to a node."""

    pass


class NodeLoad(NodeBoundaryCondition):
    """Class for adding a load to a node."""

    pass


class LineBoundaryCondition(BoundaryCondition):
    """Abstract base class for a boundary condition along a line."""

    pass



"""Library of commonly used geometries."""

from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np
from shapely import Polygon

from planestress.pre.geometry import Geometry
from planestress.pre.material import DEFAULT_MATERIAL


if TYPE_CHECKING:
    from planestress.pre.material import Material


def rectangle(
    d: float,
    b: float,
    material: Material = DEFAULT_MATERIAL,
    tol: int = 12,
) -> Geometry:
    """Creates a rectangular geometry."""
    shell = [(0, 0), (b, 0), (b, d), (0, d)]
    poly = Polygon(shell=shell)

    return Geometry(polygons=poly, materials=[material], tol=tol)


def circle(
    d: float,
    n: int,
    material: Material = DEFAULT_MATERIAL,
    tol: int = 12,
) -> Geometry:
    """Creates a circular geometry."""
    points = []

    # loop through each point on the circle
    for idx in range(n):
        # determine polar angle
        theta = idx * 2 * np.pi * 1.0 / n

        # calculate location of the point
        x = 0.5 * d * np.cos(theta)
        y = 0.5 * d * np.sin(theta)

        # append the current point to the points list
        points.append((x, y))

    poly = Polygon(shell=points)

    return Geometry(polygons=poly, materials=[material], tol=tol)



"""Class describing a planestress mesh."""

from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING, Any

import numpy as np
import numpy.typing as npt
from matplotlib.colors import ListedColormap
from matplotlib.patches import Patch
from matplotlib.tri import Triangulation
from shapely import MultiPoint, Point
from shapely.ops import nearest_points

from planestress.post.plotting import plotting_context


if TYPE_CHECKING:
    import matplotlib.axes

    from planestress.pre.material import Material


@dataclass
class Mesh:
    """Class for a plane-stress mesh."""

    nodes: npt.NDArray[np.float64]
    elements: npt.NDArray[np.float64]
    attributes: list[int]
    node_markers: list[int]
    linear: bool

    def __post_init__(self) -> None:
        """Mesh post_init method."""
        self.multi_point = MultiPoint(points=self.nodes)
        self.node_list = self.nodes.tolist()

    def num_nodes(self) -> int:
        """Returns the number of nodes in the mesh."""
        return len(self.nodes)

    def get_node(
        self,
        x: float,
        y: float,
    ) -> int:
        """Returns the node index at the point (``x``, ``y``)."""
        try:
            return self.node_list.index([x, y])
        except ValueError as exc:
            raise ValueError(f"Cannot find node at x: {x}, y: {y}.") from exc

    def get_nearest_node(
        self,
        x: float,
        y: float,
    ) -> tuple[int, tuple[float, float]]:
        """Returns the nearest node index & coordinates to the point (``x``, ``y``)."""
        nd, _ = nearest_points(g1=self.multi_point, g2=Point(x, y))
        return self.node_list.index([nd.x, nd.y]), (nd.x, nd.y)

    def plot_mesh(
        self,
        material_list: list[Material],
        nodes: bool,
        nd_num: bool,
        el_num: bool,
        materials: bool,
        alpha: float,
        mask: list[bool] | None,
        title: str,
        ux: npt.NDArray[np.float64] | None = None,
        uy: npt.NDArray[np.float64] | None = None,
        **kwargs: Any,
    ) -> matplotlib.axes.Axes:
        """Plots the finite element mesh."""
        # create plot and setup the plot
        with plotting_context(title=title, **kwargs) as (_, ax):
            assert ax

            # add deformed shape
            if ux is not None:
                x = self.nodes[:, 0] + ux
            else:
                x = self.nodes[:, 0]

            if uy is not None:
                y = self.nodes[:, 1] + uy
            else:
                y = self.nodes[:, 1]

            # create mesh triangulation (add deformed shape)
            triang = Triangulation(x, y, self.elements[:, 0:3], mask=mask)

            # if displaying materials
            if materials:
                colour_array = []
                legend_labels = []
                c = []  # Indices of elements for mapping colors

                # create an array of finite element colors
                for idx, attr in enumerate(self.attributes):
                    colour_array.append(material_list[int(attr)].colour)
                    c.append(idx)

                # create a list of unique material legend entries
                for idx, material in enumerate(material_list):
                    # if the material has not be entered yet
                    if idx == 0 or material not in material_list[0:idx]:
                        # add the material color and name to the legend list
                        patch = Patch(color=material.colour, label=material.name)
                        legend_labels.append(patch)

                cmap = ListedColormap(colors=colour_array)  # custom colormap

                # plot the mesh colors
                ax.tripcolor(
                    triang,
                    c,
                    cmap=cmap,
                )

                # display the legend
                ax.legend(
                    loc="center left",
                    bbox_to_anchor=(1, 0.5),
                    handles=legend_labels,
                )

            # plot the mesh
            ax.triplot(
                triang,
                "ko-" if nodes else "k-",
                lw=0.5,
                alpha=alpha,
            )

            # node numbers
            if nd_num:
                for idx, pt in enumerate(self.nodes):
                    ax.annotate(str(idx), xy=(pt[0], pt[1]), color="r")

            # element numbers
            if el_num:
                for idx, el in enumerate(self.elements):
                    pt1 = self.nodes[el[0]]
                    pt2 = self.nodes[el[1]]
                    pt3 = self.nodes[el[2]]
                    x = (pt1[0] + pt2[0] + pt3[0]) / 3
                    y = (pt1[1] + pt2[1] + pt3[1]) / 3
                    ax.annotate(str(idx), xy=(x, y), color="b")

        return ax



"""Methods used for solving linear systems and displaying info on tasks."""

from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np
import numpy.typing as npt
from scipy.sparse.linalg import spsolve


if TYPE_CHECKING:
    from scipy.sparse import csc_matrix


def solve_direct(
    k: npt.NDArray[np.float64],
    f: npt.NDArray[np.float64],
) -> npt.NDArray[np.float64]:
    """Solves a linear system using the direct solver method.

    Args:
        k: ``N x N`` matrix of the linear system
        f: ``N x 1`` right hand side of the linear system

    Returns:
        The solution vector to the linear system of equations
    """
    return np.linalg.solve(a=k, b=f)


def solve_direct_sparse(
    k: csc_matrix,
    f: npt.NDArray[np.float64],
) -> npt.NDArray[np.float64]:
    """Solves a sparse linear system using the direct solver method.

    Args:
        k: ``N x N`` sparse matrix of the linear system
        f: ``N x 1`` right hand side of the linear system

    Returns:
        The solution vector to the sparse linear system of equations
    """
    return spsolve(A=k, b=f)



"""Class describing a planestress material."""

from __future__ import annotations

from dataclasses import dataclass

import numpy as np
import numpy.typing as npt


@dataclass(eq=True)
class Material:
    """Class for a plane-stress material."""

    name: str = "Default"
    elastic_modulus: float = 1.0
    poissons_ratio: float = 0.0
    thickness: float = 1.0
    density: float = 1.0
    colour: str = "w"

    @property
    def mu(self) -> float:
        """Returns Lame parameter mu."""
        return self.elastic_modulus / (2 * (1 + self.poissons_ratio))

    @property
    def lda(self) -> float:
        """Returns the Lame parameter lambda."""
        return (
            self.poissons_ratio
            * self.elastic_modulus
            / ((1 + self.poissons_ratio) * (1 - 2 * self.poissons_ratio))
        )

    def get_d_matrix(
        self,
    ) -> npt.NDArray:
        """Returns the D (constitutive) matrix for plane-stress.

        sig = D . eps

        Cache this!
        """
        mu = self.mu
        lda = self.lda

        d_mat = np.zeros((3, 3))  # allocate D matrix
        d_mat[0:2, 0:2] = lda + 2 * mu
        d_mat[2, 2] = mu
        d_mat[0, 1] = lda
        d_mat[1, 0] = lda

        return d_mat


DEFAULT_MATERIAL = Material()



"""Class for a planestress analysis."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any

import matplotlib
import numpy as np
from matplotlib.colors import CenteredNorm
from matplotlib.tri import Triangulation
from mpl_toolkits.axes_grid1.axes_divider import make_axes_locatable

import planestress.analysis.solver as solver
from planestress.analysis.finite_element import FiniteElement, Tri3, Tri6
from planestress.post.plotting import plotting_context
from planestress.post.results import Results


if TYPE_CHECKING:
    import matplotlib.axes

    from planestress.pre.geometry import Geometry
    from planestress.pre.load_case import LoadCase
    from planestress.pre.mesh import Mesh


class PlaneStress:
    """Class for a plane-stress analysis."""

    def __init__(
        self,
        geometry: Geometry,
        load_cases: list[LoadCase],
        int_points: int = 3,
    ) -> None:
        """Inits the PlaneStress class."""
        self.geometry = geometry
        self.load_cases = load_cases
        self.int_points = int_points

        # initialise other class variables
        self.elements: list[FiniteElement] = []

        # check mesh has been created
        if self.geometry.mesh is None:
            raise RuntimeError(
                "No mesh detected, run Geometry.create_mesh() before creating a "
                "PlaneStress object."
            )

        self.mesh: Mesh = self.geometry.mesh

        # get finite element type
        el_type: type = Tri3 if self.mesh.linear else Tri6

        # loop through each element in the mesh
        for idx, node_idxs in enumerate(self.mesh.elements):
            # create a list containing the vertex and mid-node coordinates
            coords = self.mesh.nodes[node_idxs, :].transpose()

            # get attribute index of current element
            att_el = self.mesh.attributes[idx]

            # fetch the material
            material = self.geometry.materials[att_el]

            # add element to the list of elements
            self.elements.append(
                el_type(
                    el_idx=idx, coords=coords, node_idxs=node_idxs, material=material
                )
            )

    @staticmethod
    def dof_map(node_idxs: list[int]) -> list[int]:
        """Maps a list of node indices to a list of degrees of freedom."""
        dofs = []

        for node_idx in node_idxs:
            dofs.extend([node_idx * 2, node_idx * 2 + 1])

        return dofs

    def solve(self) -> list[Results]:
        """Solves each load case."""
        # get number of degrees of freedom
        num_dofs = self.mesh.num_nodes() * 2

        # allocate stiffness matrix and load vector
        k = np.zeros((num_dofs, num_dofs))
        k_mod = np.zeros((num_dofs, num_dofs))
        f = np.zeros(num_dofs)

        # allocate results
        results: list[Results] = []

        # assemble stiffness matrix
        for el in self.elements:
            # get element stiffness matrix
            k_el = el.element_stiffness_matrix(n_points=self.int_points)

            # get element degrees of freedom
            el_dofs = self.dof_map(node_idxs=el.node_idxs)
            el_dofs_mat = np.ix_(el_dofs, el_dofs)

            # add element stiffness matrix to global stiffness matrix
            k[el_dofs_mat] += k_el

        # for each load case
        for lc in self.load_cases:
            # assemble load vector
            for el in self.elements:
                # get element load vector
                f_el = el.element_load_vector(n_points=self.int_points)

                # get element degrees of freedom
                el_dofs = self.dof_map(node_idxs=el.node_idxs)

                # add element load vector to global load vector
                f[el_dofs] += f_el

            # apply boundary conditions
            for idx, bc in enumerate(lc.boundary_conditions):
                # get node index of current boundary condition - TODO: for segments?
                node_idx = self.mesh.node_markers.index(bc.marker_id)

                # get degrees of freedom for node index
                dofs = self.dof_map(node_idxs=[node_idx])

                # if first boundary condition get unmodified k
                k_mod = k if idx == 0 else k_mod
                k_mod, f = bc.apply_bc(k=k_mod, f=f, dofs=dofs)

            # solve system
            u = solver.solve_direct(k=k, f=f)

            # add to results
            results.append(Results(u=u))

        return results

    def plot_displacement_contour(
        self,
        results: Results,
        direction: str,
        title: str | None = None,
        cmap: str = "coolwarm",
        normalize: bool = True,
        fmt: str = "{x:.4e}",
        colorbar_label: str = "Displacement",
        alpha: float = 0.2,
        **kwargs: Any,
    ) -> matplotlib.axes.Axes:
        """Plots the displacement contours."""
        # get displacement values
        if direction == "x":
            u = results.u[0::2]
        elif direction == "y":
            u = results.u[1::2]
        elif direction == "xy":
            u = (results.u[0::2] ** 2 + results.u[1::2] ** 2) ** 0.5
        else:
            raise ValueError(f"direction must be 'x', 'y' or 'xy', not {direction}.")

        # apply title
        if not title:
            title = f"Displacement Contours [{direction}]"

        # create plot and setup the plot
        with plotting_context(title=title, **kwargs) as (fig, ax):
            assert ax

            # set up the colormap
            colormap = matplotlib.colormaps.get_cmap(cmap=cmap)

            # create triangulation
            triang = Triangulation(
                self.mesh.nodes[:, 0],
                self.mesh.nodes[:, 1],
                self.mesh.elements[:, 0:3],
            )

            if normalize:
                norm = CenteredNorm()
            else:
                norm = None

            trictr = ax.tricontourf(triang, u, cmap=colormap, norm=norm)

            # display the colorbar
            divider = make_axes_locatable(axes=ax)
            cax = divider.append_axes(position="right", size="5%", pad=0.1)

            fig.colorbar(
                mappable=trictr,
                label=colorbar_label,
                format=fmt,
                cax=cax,
            )

            # plot the finite element mesh
            self.mesh.plot_mesh(
                material_list=self.geometry.materials,
                nodes=False,
                nd_num=False,
                el_num=False,
                materials=False,
                mask=None,
                alpha=alpha,
                title=title,
                **dict(kwargs, ax=ax),
            )

        return ax

    def plot_deformed_shape(
        self,
        results: Results,
        displacement_scale: float,
        title: str | None = None,
        alpha: float = 0.2,
        **kwargs: Any,
    ) -> matplotlib.axes.Axes:
        """Plots the deformed shape."""
        # apply title
        if not title:
            title = f"Deformed Shape [ds = {displacement_scale}]"

        return self.mesh.plot_mesh(
            material_list=self.geometry.materials,
            nodes=False,
            nd_num=False,
            el_num=False,
            materials=False,
            mask=None,
            alpha=alpha,
            title=title,
            ux=results.u[0::2] * displacement_scale,
            uy=results.u[1::2] * displacement_scale,
            **kwargs,
        )
